package ru.dbotthepony.kstarbound.world.entities

import org.apache.logging.log4j.LogManager
import ru.dbotthepony.kstarbound.math.AABB
import ru.dbotthepony.kstarbound.math.Vector2d
import ru.dbotthepony.kstarbound.math.lerp
import ru.dbotthepony.kstarbound.world.Chunk
import ru.dbotthepony.kstarbound.world.ChunkPos
import ru.dbotthepony.kstarbound.world.World

/**
 * Интерфейс служит лишь для убирания жёсткой зависимости от класса Entity
 */
interface IEntity {
	val world: World<*, *>
	var chunk: Chunk<*, *>?
	var pos: Vector2d
	var rotation: Double
	val movement: MovementController<*>
	val isSpawned: Boolean
	val isRemoved: Boolean

	fun spawn()
	fun remove()
	fun think(delta: Double)
	fun onTouchSurface(velocity: Vector2d, normal: Vector2d)
}

/**
 * Определяет из себя сущность в мире, которая имеет позицию, скорость и коробку столкновений
 */
abstract class Entity(override val world: World<*, *>) : IEntity {
	override var chunk: Chunk<*, *>? = null
		set(value) {
			if (!isSpawned) {
				throw IllegalStateException("Trying to set chunk this entity belong to before spawning in world")
			}

			if (value == field) {
				return
			}

			val chunkPos = ChunkPos.fromTilePosition(pos)

			if (value != null && chunkPos != value.pos) {
				throw IllegalStateException("Set proper position before setting chunk this Entity belongs to")
			}

			val oldChunk = field
			field = value

			if (oldChunk == null && value != null) {
				world.orphanedEntities.remove(this)
				value.addEntity(this)
			} else if (oldChunk != null && value == null) {
				world.orphanedEntities.add(this)
				oldChunk.removeEntity(this)
			} else if (oldChunk != null && value != null) {
				value.transferEntity(this, oldChunk)
			}
		}

	override var pos = Vector2d()
		set(value) {
			if (field == value)
				return

			val old = field
			field = value

			if (isSpawned) {
				val oldChunkPos = ChunkPos.fromTilePosition(old)
				val newChunkPos = ChunkPos.fromTilePosition(value)

				if (oldChunkPos != newChunkPos) {
					chunk = world.getChunk(newChunkPos)?.chunk
				}
			}
		}

	override var rotation: Double = 0.0

	final override var isSpawned = false
		private set
	final override var isRemoved = false
		private set

	override fun spawn() {
		if (isSpawned)
			throw IllegalStateException("Already spawned")

		isSpawned = true
		world.entities.add(this)
		chunk = world.getChunk(ChunkPos.fromTilePosition(pos))?.chunk

		if (chunk == null) {
			world.orphanedEntities.add(this)
		}
	}

	override fun remove() {
		if (isRemoved)
			throw IllegalStateException("Already removed")

		isRemoved = true

		if (isSpawned) {
			world.entities.remove(this)
			chunk?.removeEntity(this)
		}
	}

	/**
	 * Контроллер перемещения данной сущности
	 */
	abstract override val movement: MovementController<*>
	protected abstract fun thinkAI(delta: Double)

	/**
	 * Заставляет сущность "думать".
	 */
	final override fun think(delta: Double) {
		if (!isSpawned) {
			throw IllegalStateException("Tried to think before spawning in world")
		}

		movement.thinkPhysics(delta)
		thinkAI(delta)
	}

	override fun onTouchSurface(velocity: Vector2d, normal: Vector2d) {

	}
}
